# Copyright (c) 2010-2011 Outpost Embedded, LLC
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.  You should have
# received a copy of the GNU General Public License along with this program; if
# not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA. Any Outpost Embedded trademarks that are
# incorporated in the source code or documentation are not subject to the GNU
# General Public License and may only be used or replicated with the express
# permission of Outpost Embedded, LLC.

import os, string, errno, logging
from subprocess import Popen, PIPE, STDOUT
from functools import wraps

from PyQt4 import QtCore
from pythoncom import CoInitialize, CoUninitialize
from wmi import WMI
from fdisk.device import Device
from fdisk.mbr import MBRDisk

from rapidwriter.backends.base import BaseDisk, BaseBackend
from rapidwriter.util import (
  QThreadWithEventLoop,
  locate_data_file,
  FunctionObject,
)
from rapidwriter.exceptions import (
  DeviceError,
  DeviceNotFoundError,
  DeviceBusyError,
  SubprocessError,
)


logger = logging.getLogger()


STANDARD_SUFFIX = u'USB Device'


class WindowsDisk(BaseDisk):
    wmi_object_path = None

    def __init__(self, wmi_object_path):
        super(WindowsDisk, self).__init__()
        self.wmi_object_path = wmi_object_path

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self.wmi_object_path)

    def has_wmi_object(self, wmi_object_path):
        # Windows does some funny business with some events where the object
        # path has lowercase parts and as a result can't be used for comparison
        # directly.
        return (wmi_object_path.upper() == self.wmi_object_path.upper())

    ###

    def get_name(self):
        return u'PhysicalDrive%u' % self.get_wmi_disk_drive().Index

    def get_description(self):
        s = self.get_wmi_disk_drive().Caption
        if s.endswith(STANDARD_SUFFIX):
            s = s[:-len(STANDARD_SUFFIX)].strip()
        return s

    def get_size(self):
        return long(self.get_wmi_disk_drive().Size)

    def is_removable(self):
        capabilities = self.get_wmi_disk_drive().capabilities
        if not capabilities:
            # capabilities is None for some devices on some machines.
            return False
        return (7 in capabilities)

    def get_device_path(self):
        return self.get_wmi_disk_drive().DeviceID

    ###

    def get_wmi_disk_drive(self):
        # Note: We refetch the object every time it is needed so that we can be
        # lazy and avoid dealing with marshalling COM objects between threads.
        return WMI(find_classes = False).get(self.wmi_object_path)

    def get_wmi_partition(self):
        return self.get_wmi_disk_drive().associators(
          wmi_result_class = 'Win32_DiskPartition',
        )[0]

    def get_wmi_logical_disk(self):
        return self.get_wmi_partition().associators(
          wmi_result_class = 'Win32_LogicalDisk',
        )[0]

    def get_volume(self):
        # FIXME: Volume name? (\\?\Volume{UUID})
        device_id = self.get_wmi_logical_disk().DeviceID
        drive_letter = device_id[0]
        if drive_letter not in string.uppercase:
            raise AssertionError(device_id)
        return '%s:' % drive_letter

    def _exec(self, args):
        popen = Popen(args, stdout = PIPE, stderr = STDOUT, shell = True)
        try:
            popen.wait()
            output = popen.stdout.read()
        finally:
            popen.stdout.close()
        if popen.returncode != 0:
            logger.error(output)
            raise SubprocessError(popen.returncode, output)

    ###

    def create_partition_table(self):
        try:
            device = Device(self.get_device_path())
        except IOError, e:
            logger.exception('error opening device')

            msg = unicode(str(e), errors = 'replace')
            if e.errno == errno.ENOENT:
                raise DeviceNotFoundError(msg)
            raise DeviceError(msg)

        try:
            sector_count = device.get_sector_count()

            # Force 255 heads, 63 sectors no matter what the current partition
            # table looks like.  I'm not sure if this is a good idea or not.
            # We'll see.
            geometry = MBRDisk.calculate_geometry(sector_count, 255, 63)
            mbr_disk = MBRDisk(geometry)

            with open(locate_data_file('mbr.bin', ('arch', 'pkg')), 'rb') as f:
                mbr_disk.boot_record.code.read_from_file(f)

            mbr_disk.add_partition(type = 0x0b)

            try:
                mbr_disk.write_to_device(device)
            except IOError, e:
                logger.exception('error writing to device')

                msg = unicode(str(e), errors = 'replace')
                if e.errno == errno.EBUSY:
                    raise DeviceBusyError(msg)
                raise DeviceError(msg)

        finally:
            device.close()

    def create_filesystem(self):
        # FIXME: Maybe we can do this via WMI?
        # Note: Must use shell on win32 to perform PATH searching.
        # See CreateProcess documentation for details.
        self._exec(['FORMAT', self.get_volume(), '/FS:FAT32', '/X', '/Y', '/Q'])

    def install_bootloader(self):
        syslinux = locate_data_file('syslinux.exe', ('arch', 'pkg'))
        self._exec([syslinux, self.get_volume()])

    def mount(self):
        mount_point = os.path.join(self.get_volume(), '\\')
        # All we need to do is access the volume and it will be mounted for us.
        os.listdir(mount_point)
        return mount_point

    def umount(self, mount_point):
        # This could be done via a Windows DeviceIoControl call.
        pass

    def sync(self):
        # A lock/unlock cycle will cause the filesystem to be dismounted, which
        # should flush any buffers.  A single dismount call would probably be
        # better, but it's not clear that that would force all caches to flush,
        # so Windows users should safely remove the device anyway.
        device = Device(self.get_device_path())
        try:
            device.lock()
        except IOError:
            # This is probably an "access denied" error due to some program
            # accessing the device and preventing us from locking it.
            logger.exception('error locking device')
        else:
            device.unlock()


class WMIObjectMonitor(QtCore.QObject):
    cls_name = None
    where = None

    # Note: These signals pass the WMI object path so that we can be lazy and
    # avoid dealing with marshalling COM objects between threads.

    object_created = QtCore.pyqtSignal(str)
    object_deleted = QtCore.pyqtSignal(str)

    def __init__(self, cls_name, **where):
        super(WMIObjectMonitor, self).__init__()
        self.cls_name = cls_name
        self.where = where

    @QtCore.pyqtSlot()
    def run(self):
        CoInitialize()
        try:
            wmi_conn = WMI(find_classes = False)
            cls = getattr(wmi_conn, self.cls_name)
            watcher = cls.watch_for(**self.where)
            while True:
                event = watcher()
                logger.debug('WMIObjectMonitor: got event %r', event)
                if event.event_type == 'creation':
                    self.object_created.emit(str(event.path()))
                elif event.event_type == 'deletion':
                    self.object_deleted.emit(str(event.path()))
        finally:
            CoUninitialize()


class WindowsBackend(BaseBackend):
    disks = None

    _wmi_conn = None
    _thread = None
    _monitor = None

    def __init__(self, *args, **kwargs):
        super(WindowsBackend, self).__init__(*args, **kwargs)

        self.disks = []

        self._wmi_conn = WMI(find_classes = False)

        self._thread = QThreadWithEventLoop()
        self._monitor = WMIObjectMonitor('Win32_DiskDrive')
        self._monitor.moveToThread(self._thread)
        self._thread.started.connect(self._monitor.run)
        self._monitor.object_created.connect(self._add_disk_drive)
        self._monitor.object_deleted.connect(self._remove_disk_drive)

    def start(self):
        self._thread.start()
        FunctionObject.call_in_main_thread(self._init_disks)

    def _init_disks(self):
        for disk_drive in self._wmi_conn.Win32_DiskDrive():
            self._add_disk_drive(str(disk_drive.path()))

    def _add_disk_drive(self, wmi_object_path):
        if isinstance(wmi_object_path, QtCore.QString):
            wmi_object_path = str(wmi_object_path)

        try:
            self._find_disk(wmi_object_path)
        except DeviceNotFoundError:
            logger.info('wmi: adding device: %r', wmi_object_path)
            disk = WindowsDisk(wmi_object_path)
            self.disks.append(disk)
            self.add_disk(disk)
        else:
            logger.info(
              'wmi: device already exists, not adding: %r',
              wmi_object_path,
            )

    def _remove_disk_drive(self, wmi_object_path):
        if isinstance(wmi_object_path, QtCore.QString):
            wmi_object_path = str(wmi_object_path)

        try:
            disk = self._find_disk(wmi_object_path)
        except DeviceNotFoundError:
            logger.info(
              'wmi: device not found, not removing: %r',
              wmi_object_path,
            )
        else:
            logger.info('wmi: removing device: %r', wmi_object_path)
            self.disks.remove(disk)
            # FIXME: We should probably invalidate the disk object somehow,
            # right?  I guess things will fail when the next WMI lookup is
            # attempted.
            self.remove_disk(disk)

    def _find_disk(self, wmi_object_path):
        for disk in self.disks:
            if disk.has_wmi_object(wmi_object_path):
                return disk
        raise DeviceNotFoundError()

    def process_thread(self, thread):
        thread.run = self._wrap_thread_run(thread.run)

    def _wrap_thread_run(self, fn):
        @wraps(fn)
        def wrapped(*args, **kwargs):
            CoInitialize()
            try:
                return fn(*args, **kwargs)
            finally:
                CoUninitialize()
        return wrapped
