# Copyright (c) 2010 Outpost Embedded, LLC
#
# This copyrighted material is made available to anyone wishing to use, modify,
# copy, or redistribute it subject to the terms and conditions of the GNU
# General Public License v.2.  This program is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY expressed or implied, including the
# implied warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.  You should have
# received a copy of the GNU General Public License along with this program; if
# not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
# Floor, Boston, MA 02110-1301, USA. Any Outpost Embedded trademarks that are
# incorporated in the source code or documentation are not subject to the GNU
# General Public License and may only be used or replicated with the express
# permission of Outpost Embedded, LLC.

import os, tarfile, logging

from tasks import Task, ComplexTask


logger = logging.getLogger()


CHUNK_SIZE = 10**6 # 1 MB


class DiskSubtask(Task):
    disk = None

    _cancel = None

    def __init__(self, disk, **kwargs):
        self.disk = disk
        super(DiskSubtask, self).__init__(**kwargs)

    def do_run(self):
        if self._cancel:
            return

        try:
            for task_changes in self._do_run():
                yield task_changes
        except KeyboardInterrupt:
            raise
        except Exception, e:
            logger.exception('Failing task due to exception: %r', self)
            # str(...) gives a better message for IOErrors.
            if isinstance(e, IOError):
                msg = unicode(str(e), errors = 'replace')
            else:
                msg = unicode(e)
            yield self.fail(msg)

    def do_cancel(self):
        yield [self.change('status', 'Aborting...')]
        self._cancel = True
        for task_changes in super(DiskSubtask, self).do_cancel():
            yield task_changes


class CreatePartitionTableTask(DiskSubtask):
    def __init__(self, disk, **kwargs):
        kwargs.setdefault('name', 'CreatePartitionTable')
        kwargs.setdefault('description', 'Create partition table')
        super(CreatePartitionTableTask, self).__init__(disk, **kwargs)

    def _do_run(self):
        yield [self.change('status', 'Creating partition table...')]
        logger.info('BEGIN create_partition_table')
        self.disk.create_partition_table()
        logger.info('END create_partition_table')
        yield self.complete()


class CreateFilesystemTask(DiskSubtask):
    def __init__(self, disk, **kwargs):
        kwargs.setdefault('name', 'CreateFilesystem')
        kwargs.setdefault('description', 'Create filesystem')
        super(CreateFilesystemTask, self).__init__(disk, **kwargs)

    def _do_run(self):
        yield [self.change('status', 'Creating filesystem...')]
        logger.info('BEGIN create_filesystem')
        self.disk.create_filesystem()
        logger.info('END create_filesystem')
        yield self.complete()


class InstallBootloaderTask(DiskSubtask):
    def __init__(self, disk, **kwargs):
        kwargs.setdefault('name', 'InstallBootloader')
        kwargs.setdefault('description', 'Install bootloader')
        super(InstallBootloaderTask, self).__init__(disk, **kwargs)

    def _do_run(self):
        yield [self.change('status', 'Installing bootloader...')]
        logger.info('BEGIN install_bootloader')
        self.disk.install_bootloader()
        logger.info('END install_bootloader')
        yield self.complete()


class ExtractFilesTask(DiskSubtask):
    def __init__(self, disk, source_file, **kwargs):
        kwargs.setdefault('name', 'ExtractFiles')
        kwargs.setdefault('description', 'Extract files')
        super(ExtractFilesTask, self).__init__(disk, **kwargs)
        self.source_file = source_file

    def _do_run(self):
        yield [self.change('status', 'Extracting files...')]
        logger.info('BEGIN mount')
        mount_point = self.disk.mount()
        try:
            logger.info('END mount')
            logger.info('BEGIN extract')
            logger.debug('mount_point: %r', mount_point)
            f_tar = tarfile.open(self.source_file)
            try:
                bytes_total = 0
                for member in f_tar.getmembers():
                    bytes_total += member.size

                bytes_written = 0
                for member in f_tar.getmembers():
                    dst = os.path.join(mount_point, member.name)
                    if member.isdir():
                        logger.debug('creating directory %r', dst)
                        os.mkdir(dst)
                    elif member.isfile():
                        logger.debug('creating file %r', dst)
                        f_member = f_tar.extractfile(member)
                        with open(dst, 'wb') as f:
                            while True:
                                # Make sure we don't replace the "Aborting..."
                                # status.
                                if self._cancel:
                                    break

                                # FIXME: Does os.path.basename do the right
                                # thing on Windows?  I.e. do tar member
                                # names always use forward slashes?  Can
                                # os.path.basename deal with that correctly
                                # on Windows?
                                yield [self.change(
                                  'status',
                                  (
                                    u'Extracting %s...'
                                    % os.path.basename(f_member.name)
                                  ),
                                )]

                                # Maybe cancel before I/O.
                                if self._cancel:
                                    break

                                data = f_member.read(CHUNK_SIZE)

                                if not data:
                                    break

                                # Maybe cancel before I/O.
                                if self._cancel:
                                    break

                                f.write(data)
                                bytes_written += len(data)

                                progress = int(
                                  (float(bytes_written) * (100.0))
                                  / float(bytes_total)
                                )
                                yield [self.change('progress', progress)]

            finally:
                f_tar.close()
            logger.info('END extract')
        finally:
            logger.info('BEGIN umount')
            self.disk.umount(mount_point)
            logger.info('END umount')
        yield self.complete()


class SyncTask(DiskSubtask):
    def __init__(self, disk, **kwargs):
        kwargs.setdefault('name', 'Sync')
        kwargs.setdefault('description', 'Flush data to disk')
        super(SyncTask, self).__init__(disk, **kwargs)

    def _do_run(self):
        yield [self.change('status', 'Flushing data to disk...')]
        logger.info('BEGIN sync')
        self.disk.sync()
        logger.info('END sync')
        yield self.complete()


class PrepareDiskTask(ComplexTask):
    def __init__(self, disk, source_file):
        self.disk = disk
        super(PrepareDiskTask, self).__init__(
          name = 'PrepareDisk',
          description = 'Prepare disk',
          subtasks = [
            CreatePartitionTableTask(disk, progress_weight = 1),
            CreateFilesystemTask(disk, progress_weight = 2),
            InstallBootloaderTask(disk, progress_weight = 1),
            ExtractFilesTask(disk, source_file, progress_weight = 20),
            SyncTask(disk, progress_weight = 2),
          ],
        )

    def do_run(self):
        self.disk.start()
        try:
            for task_changes in super(PrepareDiskTask, self).do_run():
                yield task_changes
        finally:
            self.disk.stop()
